from octostar.client import make_client, UserContext, User
from functools import wraps
from octostar_streamlit.desktop import whoami
from streamlit.runtime.scriptrunner import get_script_run_ctx
import hashlib
import streamlit as st
import time


def impersonating_running_user(**client_kwargs):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            client = as_running_user(**client_kwargs).user.client
            kwargs["client"] = client
            return func(*args, **kwargs)

        return wrapper

    return decorator


def as_running_user(force_refresh=False):
    session = st.session_state.setdefault("__run_user", {})
    script_ctx = get_script_run_ctx()
    current_time = time.time()

    should_refresh = (
        force_refresh
        or script_ctx.script_requests is not session.get("prev_run")
        or current_time > session.get("token_expiry", 0) - 300  # 5 minutes buffer
    )

    if not should_refresh and "prev_user" in session:
        return UserContext(session["prev_user"])

    running_user = whoami()
    if not running_user:
        return (
            UserContext(session.get("prev_user"))
            if "prev_user" in session
            else st.stop()
        )

    user_hash = int(hashlib.md5(running_user.os_jwt.encode("utf-8")).hexdigest(), 16)
    if "prev_user" not in session or hash(session["prev_user"]) != user_hash:
        client = make_client(fixed_jwt=running_user.os_jwt)
        user = User(client)
        session["prev_user"] = user
        session["token_expiry"] = user.jwt_expires_at

    session["prev_run"] = script_ctx.script_requests
    return UserContext(session["prev_user"])



import threading
import time
import streamlit as st
import uuid
from streamlit.runtime.scriptrunner import get_script_run_ctx
from streamlit.runtime import get_instance as get_runtime_instance

from .session_state_hot_swapper import SessionStateHotSwapper


class SessionCallbackManager(object):
    instance = None

    def __new__(cls, *args, **kwargs):
        if not cls.instance:
            cls.instance = super().__new__(cls)
            cls.instance.initialized = False
        return cls.instance

    def __init__(self):
        if not self.initialized:
            self.initialized = True
            self.active_sessions = {}
            self.monitoring_thread_active = False
            self.monitoring_thread_lock = threading.Lock()

    @st.cache_resource
    def get_handler():
        if not SessionCallbackManager.instance:
            SessionCallbackManager.instance = SessionCallbackManager()
        return SessionCallbackManager.instance

    def heartbeat(self):
        inactive_sessions = []
        for session_id, session_info in list(self.active_sessions.items()):
            if session_info["runtime"].is_active_session(
                session_id=session_info["context"].session_id
            ):
                heartbeat_fn = session_info.get("on_heartbeat")
                heartbeat_fn(session_info["context"])
            else:
                inactive_sessions.append(session_id)
        for session_id in inactive_sessions:
            self.on_session_end(session_id)

    def monitoring_thread(self):
        while self.monitoring_thread_active:
            self.heartbeat()
            time.sleep(2)

    def start_monitoring(self):
        with self.monitoring_thread_lock:
            if not self.monitoring_thread_active:
                self.monitoring_thread_active = True
                thread = threading.Thread(target=self.monitoring_thread)
                thread.daemon = True
                thread.start()

    def register_session(
        self, on_session_start=None, on_heartbeat=None, on_session_end=None
    ):
        session_id = str(uuid.uuid4())
        ctx = get_script_run_ctx()
        runtime = get_runtime_instance()
        with self.monitoring_thread_lock:
            self.active_sessions[session_id] = {}
            self.active_sessions[session_id]["session_id"] = session_id
            self.active_sessions[session_id]["context"] = ctx
            self.active_sessions[session_id]["runtime"] = runtime
            self.active_sessions[session_id]["on_session_start"] = (
                SessionCallbackManager.callback_wrapper(on_session_start)
            )
            self.active_sessions[session_id]["on_heartbeat"] = (
                SessionCallbackManager.callback_wrapper(on_heartbeat)
            )
            self.active_sessions[session_id]["on_session_end"] = (
                SessionCallbackManager.callback_wrapper(on_session_end)
            )
        self.on_session_start(session_id)
        return session_id

    def callback_wrapper(callback_fn):
        def _callback(st_context):
            if callback_fn:
                with SessionStateHotSwapper(st_context):
                    callback_fn()

        return _callback

    def on_session_start(self, session_id):
        start_fn = self.active_sessions[session_id].get("on_session_start")
        start_fn(self.active_sessions[session_id]["context"])
        self.start_monitoring()

    def on_session_end(self, session_id):
        with self.monitoring_thread_lock:
            if session_id in self.active_sessions:
                end_fn = self.active_sessions[session_id].get("on_session_end")
                end_fn(self.active_sessions[session_id]["context"])
                self.active_sessions[session_id]["context"].session_state._state.clear()
                self.active_sessions.pop(session_id, None)
                if not self.active_sessions:
                    self.monitoring_thread_active = False



import streamlit as st

from octostar_streamlit.desktop import get_open_workspace_ids
from octostar.utils.ontology import query_ontology
from octostar.utils.workspace.permissions import get_permissions, PermissionLevel


def get_workspaces(client, prev_workspaces=None, include_closed=False):
    if not include_closed:
        workspace_ids = get_open_workspace_ids()
    else:
        workspace_ids = [
            w["os_entity_uid"]
            for w in client.execute(
                query_ontology.sync,
                f"SELECT `os_entity_uid` FROM `dtimbr`.`os_workspace`",
            )
        ]
    if not workspace_ids and prev_workspaces is not None:
        return prev_workspaces
    if not workspace_ids:
        st.stop()
    workspace_ids = ", ".join(
        ["'" + workspace_id + "'" for workspace_id in workspace_ids]
    )
    workspaces = client.execute(
        query_ontology.sync,
        f"SELECT `entity_label`, `os_entity_uid` FROM `dtimbr`.`os_workspace` WHERE `os_entity_uid` IN ({workspace_ids})",
    )
    workspaces_permissions = client.execute(
        get_permissions.sync, [w["os_entity_uid"] for w in workspaces]
    )
    workspaces = {
        workspace["os_entity_uid"]: {
            **workspace,
            "os_permission": workspaces_permissions.get(
                workspace["os_entity_uid"], PermissionLevel.NONE
            ),
        }
        for workspace in workspaces
    }
    return dict(
        filter(
            lambda x: x[1]["os_permission"] > PermissionLevel.NONE, workspaces.items()
        )
    )



import os
import yaml
import json
import base64

from .entities_parser import EntitiesParser, EntitiesParserRuleset
from .parameters import ConsolidationParameters

# from octostar.client import impersonating_launching_user, dev_mode


# @impersonating_launching_user()
# @dev_mode(os.getenv('OS_DEV_MODE'))
def parse_root_nodes_to_linkchart(
    endpoint,
    source,
    body,
    transformResult,
    addToGraph,
    guess_entries_fn,
    guess_rulesets_fn,
    guess_entity_fn,
    get_parsed_label_fn,
    process_invalid_fn,
    ruleset_params,
    parser_name,
    client=None,
):
    # The correct ontological path should be source -> entry (api_response, non-parsed) -> account (root node, parsed) -> emails, people... (non-root nodes, parsed)
    # However, because this looks like a longer path in the linkchart and the requirement is not to while still having a two-steps procedure,
    # we need a way to replace the api_response with the parsed root.
    # Also because the linkchart will bug out by showing two nodes with the same UUID if they have different entity type
    # (the actual id is built with both but no way to externally control this),
    # the only solution (for now) is to compute the parsed nodes immediately to get the entity type and UUID they imply, but still show the raw data.
    # In the future hiding the middle node should really be a styling feature
    entries = guess_entries_fn(endpoint, transformResult)
    records = list()
    relationships = list()
    for entry in entries:
        guessed_entity = guess_entity_fn(endpoint, entry)
        ruleset_files, ruleset_parameters = guess_rulesets_fn(endpoint, entry)
        if isinstance(ruleset_params, list):
            for idx, ruleset_parameter in enumerate(ruleset_parameters):
                if ruleset_parameter and not ruleset_params[idx]:
                    pass
                else:
                    ruleset_parameters[idx] = ruleset_params[idx]
        parser = EntitiesParser(
            [
                EntitiesParserRuleset(
                    os.path.join("rulesets", ruleset_files[i]), ruleset_parameters[i]
                )
                for i in range(len(ruleset_files))
            ],
            client=client,
        )
        new_entry_entities, _ = parser.apply_rules(
            entry, endpoint, body, guessed_entity
        )
        new_entry_entities = list(filter(lambda x: not x.parents, new_entry_entities))
        processed_entities, processed_relationships = (
            parser.validate_and_format_entities(new_entry_entities, [])
        )
        processed_entities, processed_relationships = process_invalid_fn(
            processed_entities, processed_relationships
        )
        new_linkchart_elements = parser.to_linkchart(
            processed_entities, processed_relationships, get_parsed_label_fn
        )
        for new_linkchart_element in new_linkchart_elements["records"]:
            records.append(
                {
                    "#endpoint": endpoint,
                    "#original_entry": base64.b64encode(
                        json.dumps(entry).encode("utf-8")
                    ).decode("utf-8"),
                    "#parser": parser_name,
                    "#guessed_entity": guessed_entity,
                    "#is_empty": not bool(entry),
                    "#input": body,
                    "#ruleset_parameters": ruleset_parameters,
                    **new_linkchart_element,
                }
            )
            relationships.append(
                {
                    "from": source["os_entity_uid"],
                    "to": new_linkchart_element["os_entity_uid"],
                    "label": endpoint,
                }
            )
    if len(records) == 1 and ("#is_empty" not in records[0] or records[0]["#is_empty"]):
        records = list()
        relationships = list()
    addToGraph["records"].extend(records)
    addToGraph["relationships"].extend(relationships)
    addToGraph = {
        "records": addToGraph["records"],
        "relationships": addToGraph["relationships"],
    }
    return addToGraph


# @impersonating_launching_user()
# @dev_mode(os.getenv('OS_DEV_MODE'))
def parse_all_nodes_to_linkchart(
    sources,
    addToGraph,
    guess_rulesets_fn,
    get_parsed_label_fn,
    process_invalid_fn,
    client=None,
):
    with open(
        os.path.join(os.getcwd(), "rulesets", "consolidation_parameters.yaml")
    ) as file:
        consolidation_params_dict = yaml.safe_load(file)
    consolidation_parameters = ConsolidationParameters(
        equality_tolerances=consolidation_params_dict.get("tolerances", dict()).get(
            "equality", dict()
        ),
        similarity_tolerances=consolidation_params_dict.get("tolerances", dict()).get(
            "similarity", dict()
        ),
        keep_roots=True,
    )
    records = list()
    relationships = list()
    parsed_entities = list()
    parsed_relationships = list()
    for source in sources:
        endpoint = source["#endpoint"]
        guessed_entity = source["#guessed_entity"]
        body = source["#input"]
        entry = json.loads(
            base64.b64decode(source["#original_entry"].encode("utf-8")).decode("utf-8")
        )
        ruleset_files, _ = guess_rulesets_fn(endpoint, source)
        ruleset_parameters = source["#ruleset_parameters"]
        parser = EntitiesParser(
            [
                EntitiesParserRuleset(
                    os.path.join("rulesets", ruleset_files[i]), ruleset_parameters[i]
                )
                for i in range(len(ruleset_files))
            ],
            client=client,
        )
        new_entities, new_relationships = parser.apply_rules(
            entry, endpoint, body, guessed_entity
        )
        parsed_entities.extend(new_entities)
        parsed_relationships.extend(new_relationships)
    parser = EntitiesParser([])
    consolidated_entities, consolidated_relationships, _ = parser.consolidate_entities(
        parsed_entities, parsed_relationships, consolidation_parameters
    )
    processed_entities, processed_relationships = parser.validate_and_format_entities(
        consolidated_entities, consolidated_relationships
    )
    processed_entities, processed_relationships = process_invalid_fn(
        processed_entities, processed_relationships
    )
    new_linkchart_elements = parser.to_linkchart(
        processed_entities, processed_relationships, get_parsed_label_fn
    )
    for new_linkchart_element in new_linkchart_elements["records"]:
        records.append({"#is_empty": not bool(entry), **new_linkchart_element})
    relationships.extend(new_linkchart_elements["relationships"])
    if len(records) == 1 and ("#is_empty" not in records[0] or records[0]["#is_empty"]):
        records = list()
        relationships = list()
    for record in records:
        del record["#is_empty"]
    addToGraph["records"].extend(records)
    addToGraph["relationships"].extend(relationships)
    addToGraph = {
        "records": addToGraph["records"],
        "relationships": addToGraph["relationships"],
    }
    return addToGraph



import re
import py3langid as langid
import iso639 as languages


def alpha2_to_language(alpha2: str) -> str:
    if not alpha2:
        return None
    code = alpha2.strip().lower()
    return languages.to_name(code)

def language_to_alpha2(language_name: str) -> str:
    if not language_name:
        return None
    name = language_name.strip().lower()
    data = languages.find(name)
    return data["iso639_1"]

def detect_language(text, min_confidence=None):
    detector = langid.langid.LanguageIdentifier.from_pickled_model(
        langid.langid.MODEL_FILE, norm_probs=True
    )
    detected_lang, confidence = detector.classify(text)
    if min_confidence and confidence < min_confidence:
        return None, confidence
    detected_lang = re.sub("[^A-Za-z]", "", detected_lang).lower()
    detected_lang = languages.to_name(detected_lang).lower()
    return detected_lang, confidence

FLAIR_MODELS = {
    "en": "flair/ner-english-large",
    "es": "flair/ner-spanish-large",
    "de": "flair/ner-german-large",
    "nl": "flair/ner-dutch-large",
    "multi": "flair/ner-multi",
    "multi-fast": "flair/ner-multi-fast",
}

SPACY_MODELS = {
    "en": 'en_core_web_sm',
}

def load_language_model(language, type):
    from flair.models import SequenceTagger
    from spacy_download import load_spacy

    model = None
    match type:
        case "spacy":
            model_name = SPACY_MODELS.get(language, SPACY_MODELS["en"])
            model = load_spacy(model_name)
        case "flair":
            model_name = FLAIR_MODELS.get(language, "flair/ner-multi")
            model = SequenceTagger.load(model_name)
    return model


import mimetypes
import filetype


def add_filetypes():
    mimetypes.add_type("text/csv", ".csv")


def recognize_filetype(file=None, filename=None):
    if not (file or filename):
        raise ValueError("Must provide either file content or filename")
    from_name_guess = None
    from_file_guess = None
    if filename:
        from_name_guess, _ = mimetypes.guess_type(filename)
    if file:
        ft = filetype.guess(file)
        from_file_guess = ft.mime if ft else None
    def _family(m):
        return m.split("/", 1)[0] if isinstance(m, str) and "/" in m else None
    def _subtype(m):
        return m.split("/", 1)[-1] if isinstance(m, str) and "/" in m else None
    if not from_file_guess and not from_name_guess:
        return "unknown"
    if not (from_file_guess and from_name_guess):
        return from_file_guess or from_name_guess
    if _family(from_file_guess) != _family(from_name_guess):
        return from_file_guess
    if _subtype(from_file_guess) != _subtype(from_name_guess):
        return from_name_guess
    return from_name_guess



from itertools import groupby, chain
import time
import uuid
from octostar.utils.ontology import multiquery_ontology
from octostar.utils.exceptions import StopAsyncIterationWithResult

from .inheritance import is_child_concept

REL_FETCHED_FIELDS = [
    "os_relationship_name",
    "os_entity_uid_from",
    "os_entity_uid_to",
    "os_entity_uid",
    "os_entity_type_from",
    "os_entity_type_to",
    "os_workspace",
]
TARGETS_FETCHED_FIELDS = ["os_entity_uid", "os_concept", "entity_label", "os_workspace"]


class ExecutionMetrics:
    def __init__(self):
        self.n_queries = 0
        self.n_relationships = 0
        self.n_target_entities = 0
        self.exec_time = 0.0
        self.exec_times = {}
        self.n_cancelled_queries = 0
        self.relationship_names = set()
        self.timeout = 0

    def print_metrics(self):
        print("Execution Metrics:")
        print(f"  Execution Time: {self.exec_time}")
        print(f"  Per-Type Execution Times: {self.exec_times}")
        print(f"  Number of Queries: {self.n_queries}")
        print(f"  Number of Cancelled Queries: {self.n_cancelled_queries}")
        print(f"  Timeout per Query: {self.timeout}")
        print(f"  Number of Relationships: {self.n_relationships}")
        print(f"  Number of Target Entities: {self.n_target_entities}")
        print(f"  Relationship Names: {self.relationship_names}")


async def _get_stream_result(stream):
    result = None
    try:
        async for _ in stream:
            pass
    except StopAsyncIterationWithResult as e:
        result = e.value
    return result


def _left_join(left_list, right_list, left_keys, right_keys):
    left_keys = left_keys if isinstance(left_keys, (list, tuple)) else [left_keys]
    right_keys = right_keys if isinstance(right_keys, (list, tuple)) else [right_keys]

    def _make_composite_key(item, keys):
        return tuple(item[key] for key in keys)

    left_list.sort(key=lambda item: _make_composite_key(item, left_keys))
    right_list.sort(key=lambda item: _make_composite_key(item, right_keys))
    left_groups = groupby(
        left_list, key=lambda item: _make_composite_key(item, left_keys)
    )
    right_groups = groupby(
        right_list, key=lambda item: _make_composite_key(item, right_keys)
    )
    right_dict = {key: list(group) for key, group in right_groups}
    result = []
    for left_key_val, left_items in left_groups:
        left_items = list(left_items)
        associated_rights = right_dict.get(left_key_val, [])
        for left_item in left_items:
            result.append((left_item, associated_rights))
    return result


async def expand_otm(
    expanded_entities,
    entities_by_concept,
    ontology_rels,
    concepts_to_otm_rels,
    metrics,
    client,
    timeout,
    limit,
):
    start_time = time.time()
    otm_queries = []
    target_fields = ",".join(
        ["`" + rel_field + "`" for rel_field in TARGETS_FETCHED_FIELDS]
    )
    for concept_name, rels in concepts_to_otm_rels.items():
        for rel_name in rels:
            rel = ontology_rels[rel_name]
            source_properties = rel["source_properties"].split(",")
            source_properties_values = [
                [entity[p] for p in source_properties]
                for entity in entities_by_concept[concept_name]
            ]
            source_properties_values = [
                values
                for values in source_properties_values
                if all(value is not None for value in values)
            ]
            if source_properties_values:
                target_prop_names = rel["target_properties"].split(",")
                target_prop_names = ",".join(
                    ["`" + prop + "`" for prop in target_prop_names]
                )
                source_properties_values = ",".join(
                    [
                        "(" + ",".join(["'" + value + "'" for value in values]) + ")"
                        for values in source_properties_values
                    ]
                )
                if target_prop_names and source_properties_values:
                    query = f"""SELECT {target_fields} FROM `etimbr`.`{rel['target_concept']}` WHERE ({target_prop_names}) IN ({source_properties_values}) LIMIT {limit}"""
                    otm_queries.append(
                        {
                            "concept_name": concept_name,
                            "query": query,
                            "relationship_name": rel_name,
                            "source_properties": rel["source_properties"].split(","),
                            "target_properties": rel["target_properties"].split(","),
                        }
                    )
    otm_queries = {str(i): otm_queries[i] for i in range(len(otm_queries))}
    otm_stream = multiquery_ontology.streaming(
        sql_queries={
            query_id: query_data["query"]
            for query_id, query_data in otm_queries.items()
        },
        client=client,
        timeout=timeout,
    )
    metrics.n_queries += len(otm_queries)
    all_results = await _get_stream_result(otm_stream)
    for query_id, data in all_results.items():
        if data == None:
            metrics.n_cancelled_queries += 1
        if not data:
            continue
        query_data = otm_queries[query_id]
        data = _left_join(
            [
                e[0]
                for e in expanded_entities.values()
                if e[0]["os_concept"] == query_data["concept_name"]
            ],
            data,
            query_data["source_properties"],
            query_data["target_properties"],
        )
        data = {
            e[0]["os_entity_uid"]: (
                e[0],
                {
                    r["os_entity_uid"]: (
                        {"os_relationship_name": query_data["relationship_name"]},
                        {r["os_entity_uid"]: r},
                    )
                    for r in e[1]
                },
            )
            for e in data
        }
        for entity_id, entities in data.items():
            if entity_id not in expanded_entities:
                expanded_entities[entity_id] = [entities[0], {}]
            for sub_entity_id, sub_entities in entities[1].items():
                if sub_entity_id not in expanded_entities[entity_id][1]:
                    expanded_entities[entity_id][1][sub_entity_id] = [
                        sub_entities[0],
                        {},
                    ]
                expanded_entities[entity_id][1][sub_entity_id][1].update(
                    sub_entities[1]
                )
    metrics.exec_times["otm"] = time.time() - start_time
    return expanded_entities


async def expand_mtm_mixed(
    expanded_entities,
    entities_by_concept,
    ontology,
    concepts_to_mixed_rels,
    ontology_rels,
    metrics,
    client,
    timeout,
    limit,
):
    start_time = time.time()
    mixed_queries = []
    match_patterns = {}
    rel_fields = ",".join(["`" + rel_field + "`" for rel_field in REL_FETCHED_FIELDS])
    for concept_name, rels in concepts_to_mixed_rels.items():
        if concept_name not in match_patterns:
            match_patterns[concept_name] = []
        match_patterns[concept_name].extend(
            [entity["os_entity_uid"] for entity in entities_by_concept[concept_name]]
        )
    for concept_name, entity_ids in match_patterns.items():
        rels = concepts_to_mixed_rels[concept_name]
        entity_ids = ",".join(["'" + uid + "'" for uid in entity_ids])
        for rel_name in rels:
            rel = ontology_rels[rel_name]
            target_concept = rel["target_concept"]
            relationship_fields = ",".join(
                [
                    f"`{rel_name}[{target_concept}]_"
                    + prop
                    + "` AS "
                    + "`rel__"
                    + prop
                    + "`"
                    for prop in REL_FETCHED_FIELDS
                ]
            )
            target_fields = ",".join(
                [
                    f"`{rel_name}[{target_concept}]."
                    + prop
                    + "` AS "
                    + "`tgt__"
                    + prop
                    + "`"
                    for prop in TARGETS_FETCHED_FIELDS
                ]
            )
            all_fields = (
                f"{relationship_fields}, {target_fields}, `os_entity_uid`".strip(", ")
            )
            if entity_ids:
                mixed_queries.append(
                    {
                        "query": f"SELECT {all_fields} FROM `dtimbr`.`{concept_name}` WHERE `os_entity_uid` IN ({entity_ids}) LIMIT {limit}"
                    }
                )
    mixed_queries = {str(i): mixed_queries[i] for i in range(len(mixed_queries))}
    metrics.n_queries += len(mixed_queries)
    local_stream = multiquery_ontology.streaming(
        sql_queries={
            query_id: query_data["query"]
            for query_id, query_data in mixed_queries.items()
        },
        client=client,
        timeout=timeout / 2.0,
    )
    all_results = await _get_stream_result(local_stream)
    for (
        _,
        data,
    ) in (
        all_results.items()
    ):  ## TO BE TESTED (data is always empty at the moment due to timbr bug)
        if data == None:
            metrics.n_cancelled_queries += 1
        if not data:
            continue
        data = _left_join(
            [e[0] for e in expanded_entities.values()],
            data,
            "os_entity_uid",
            "os_entity_uid",
        )
        for elem in data:
            elem["#rel__os_entity_uid"] = elem["rel__os_entity_uid"] or (
                "temp-" + str(uuid.uuid4())
            )
        data = {
            e[0]["os_entity_uid"]: (
                e[0],
                {
                    rt["#rel__os_entity_uid"]: (
                        {k[5:]: v for k, v in rt.items() if k.startswith("rel__")},
                        {
                            rt["tgt__os_entity_uid"]: {
                                k[5:]: v for k, v in rt.items() if k.startswith("tgt__")
                            }
                        },
                    )
                    for rt in e[1]
                },
            )
            for e in data
        }
        data = {
            e[0]["os_entity_uid"]: (
                e[0],
                {r["os_entity_uid"]: ({}, {r["os_entity_uid"]: r}) for r in e[1]},
            )
            for e in data
        }
        for entity_id, entities in data.items():
            if entity_id not in expanded_entities:
                expanded_entities[entity_id] = [entities[0], {}]
            for sub_entity_id, sub_entities in entities[1].items():
                if sub_entity_id not in expanded_entities[entity_id][1]:
                    expanded_entities[entity_id][1][sub_entity_id] = [
                        sub_entities[0],
                        {},
                    ]
                expanded_entities[entity_id][1][sub_entity_id][1].update(
                    sub_entities[1]
                )
    metrics.exec_times["mtm_mixed"] = time.time() - start_time
    return expanded_entities


async def expand_mtm_local(
    expanded_entities,
    entities_by_concept,
    ontology_rels,
    concepts_to_local_rels,
    metrics,
    client,
    timeout,
    limit,
):
    start_time = time.time()
    local_queries = {"from": [], "to": []}
    match_patterns = {}
    target_fields = ",".join(
        ["`" + rel_field + "`" for rel_field in TARGETS_FETCHED_FIELDS]
    )
    rel_fields = ",".join(["`" + rel_field + "`" for rel_field in REL_FETCHED_FIELDS])
    for concept_name in concepts_to_local_rels.keys():
        if concept_name not in match_patterns:
            match_patterns[concept_name] = []
        match_patterns[concept_name].extend(
            [entity["os_entity_uid"] for entity in entities_by_concept[concept_name]]
        )
    for concept_name, entity_ids in match_patterns.items():
        rel_names = concepts_to_local_rels[concept_name]
        inverse_names = [
            ontology_rels[rel_name]["inverse_name"] for rel_name in rel_names
        ]
        rel_names = ",".join(["'" + rel_name + "'" for rel_name in rel_names])
        inverse_names = ",".join(["'" + rel_name + "'" for rel_name in inverse_names])
        entity_ids = ",".join(["'" + uid + "'" for uid in entity_ids])
        if rel_names and inverse_names and entity_ids:
            local_queries["from"].append(
                {
                    "query": f"SELECT {rel_fields} FROM `timbr`.`os_workspace_relationship` WHERE `os_relationship_name` IN ({rel_names}) AND `os_entity_uid_from` IN ({entity_ids}) LIMIT {limit}"
                }
            )
            local_queries["to"].append(
                {
                    "query": f"SELECT {rel_fields} FROM `timbr`.`os_workspace_relationship` WHERE `os_relationship_name` IN ({inverse_names}) AND `os_entity_uid_to` IN ({entity_ids}) LIMIT {limit}"
                }
            )
    local_queries = {
        **{
            "from_" + str(i): local_queries["from"][i]
            for i in range(len(local_queries["from"]))
        },
        **{
            "to_" + str(i): local_queries["to"][i]
            for i in range(len(local_queries["to"]))
        },
    }
    metrics.n_queries += len(local_queries)
    local_stream = multiquery_ontology.streaming(
        sql_queries={
            query_id: query_data["query"]
            for query_id, query_data in local_queries.items()
        },
        client=client,
        timeout=timeout / 2.0,
    )
    all_results = await _get_stream_result(local_stream)
    middle_entities = []
    for query_id, data in all_results.items():
        if data == None:
            metrics.n_cancelled_queries += 1
        if not data:
            continue
        query_id = query_id.split("_")
        data = _left_join(
            [e[0] for e in expanded_entities.values()],
            data,
            "os_entity_uid",
            "os_entity_uid_" + query_id[0],
        )
        inverse_direction = "to" if query_id[0] == "from" else "from"
        data = {
            e[0]["os_entity_uid"]: (e[0], {r["os_entity_uid"]: (r, {}) for r in e[1]})
            for e in data
        }
        middle_entities.extend(
            [(inverse_direction, r) for e in data.values() for r in e[1].values()]
        )
        for entity_id, entities in data.items():
            if entity_id not in expanded_entities:
                expanded_entities[entity_id] = [entities[0], {}]
            for sub_entity_id, sub_entities in entities[1].items():
                if sub_entity_id not in expanded_entities[entity_id][1]:
                    expanded_entities[entity_id][1][sub_entity_id] = [
                        sub_entities[0],
                        {},
                    ]
                expanded_entities[entity_id][1][sub_entity_id][1].update(
                    sub_entities[1]
                )
    middle_entities = [
        {
            "entity_id": r[1][0]["os_entity_uid_" + r[0]],
            "concept_name": r[1][0]["os_entity_type_" + r[0]],
            "direction": r[0],
            "relationship": r[1],
        }
        for r in middle_entities
    ]
    middle_entities = sorted(middle_entities, key=lambda x: x["concept_name"])
    middle_entities = groupby(middle_entities, key=lambda x: x["concept_name"])
    middle_entities = {e[0]: list(e[1]) for e in middle_entities}
    local_queries_2 = []
    for concept_name, entities in middle_entities.items():
        entity_ids = ",".join(["'" + e["entity_id"] + "'" for e in entities])
        if entity_ids:
            local_queries_2.append(
                {
                    "query": f"SELECT {target_fields} FROM `timbr`.`{concept_name}` WHERE `os_entity_uid` IN ({entity_ids}) LIMIT {limit}"
                }
            )
    local_queries_2 = {str(i): local_queries_2[i] for i in range(len(local_queries_2))}
    metrics.n_queries += len(local_queries_2)
    local_stream_2 = multiquery_ontology.streaming(
        sql_queries={
            query_id: query_data["query"]
            for query_id, query_data in local_queries_2.items()
        },
        client=client,
        timeout=timeout / 2.0,
    )
    all_results = await _get_stream_result(local_stream_2)
    middle_entities = list(chain(*middle_entities.values()))
    for query_id, data in all_results.items():
        if data == None:
            metrics.n_cancelled_queries += 1
        if not data:
            continue
        data = _left_join(middle_entities, data, "entity_id", "os_entity_uid")
        for entry in data:
            relationship_targets = entry[0]["relationship"]
            for target_entity in entry[1]:
                relationship_targets[1][target_entity["os_entity_uid"]] = target_entity
        if data:
            for rel in data:
                targets = rel[0]["relationship"][1]
                rel_uid = rel[0]["relationship"][0]["os_entity_uid"]
                for entity in expanded_entities.values():
                    entity_rels = entity[1]
                    if rel_uid in entity_rels:
                        entity_rels[rel_uid][1].update(targets)
    metrics.exec_times["mtm_local"] = time.time() - start_time
    return expanded_entities


async def expand_entities(
    entities,
    ontology,
    relationship_mappings_info,
    client,
    relationship_names_by_entity_type=None,
    batch_size=10000,
    avg_limit_per_entity=20,
    timeout=10.0,
):
    metrics = ExecutionMetrics()
    metrics.timeout = timeout
    start_time = time.time()
    ordered_entities = sorted(entities, key=lambda x: x["entity_type"])
    ordered_entities = list({e["os_entity_uid"]: e for e in entities}.values())
    entity_batches = [
        ordered_entities[i : i + batch_size]
        for i in range(0, len(ordered_entities), batch_size)
    ]
    expanded_entities = {}
    for batch in entity_batches:
        new_entities = await expand_entities_batch(
            batch,
            ontology,
            relationship_mappings_info,
            relationship_names_by_entity_type,
            metrics,
            client,
            avg_limit_per_entity,
            timeout,
        )
        expanded_entities.update(new_entities)
    metrics.exec_time = time.time() - start_time
    for _, entity_data in expanded_entities.items():
        for rel in entity_data[1].values():
            if rel[0]:
                metrics.n_relationships += 1
                metrics.relationship_names.add(rel[0]["os_relationship_name"])
            metrics.n_target_entities += len(rel[1])
    metrics.relationship_names = list(metrics.relationship_names)
    expanded_entities = [
        (e[0], [(r[0], list(r[1].values())[0]) for r in e[1].values()])
        for e in expanded_entities.values()
    ]
    return expanded_entities, metrics


async def expand_entities_batch(
    entities,
    ontology,
    relationship_mappings,
    relationship_names_by_entity_type,
    metrics,
    client,
    avg_limit_per_entity=20,
    timeout=10.0,
):
    limit = avg_limit_per_entity * len(entities)
    expanded_entities = {e["os_entity_uid"]: [e, {}] for e in entities}
    entities_by_concept = sorted(entities, key=lambda x: x["entity_type"])
    entities_by_concept = groupby(entities_by_concept, key=lambda x: x["entity_type"])
    entities_by_concept = {e[0]: list(e[1]) for e in entities_by_concept}
    ontology_rels = {r["relationship_name"]: r for r in ontology["relationships"]}
    concepts_to_rels = {
        cn: c["relationships"] for cn, c in ontology["concepts"].items()
    }
    concepts_to_mixed_rels = {}
    concepts_to_otm_rels = {}
    concepts_to_local_rels = {}
    for concept_name in entities_by_concept.keys():
        relationship_names = None
        if relationship_names_by_entity_type:
            relationship_names = set()
            for cn_name in relationship_names_by_entity_type.keys():
                if is_child_concept(concept_name, cn_name, ontology):
                    relationship_names = relationship_names.union(
                        set(relationship_names_by_entity_type[cn_name])
                    )
        rels = concepts_to_rels[concept_name]
        if relationship_names:
            rels = [r for r in rels if r in relationship_names]
        filtered_mtm_rels = []
        filtered_otm_rels = []
        filtered_local_mtm_rels = []
        for rel in rels:
            if not ontology_rels[rel]["is_mtm"]:
                filtered_otm_rels.append(rel)
            elif rel in relationship_mappings["unmapped"]:
                continue
            elif rel in relationship_mappings["local_only"]:
                filtered_local_mtm_rels.append(rel)
            else:
                filtered_mtm_rels.append(rel)
        concepts_to_mixed_rels[concept_name] = filtered_mtm_rels
        concepts_to_otm_rels[concept_name] = filtered_otm_rels
        concepts_to_local_rels[concept_name] = filtered_local_mtm_rels
    # OTM QUERIES: fetch target entities directly from the target tables with an IN statement
    expanded_entities = await expand_otm(
        expanded_entities,
        entities_by_concept,
        ontology_rels,
        concepts_to_otm_rels,
        metrics,
        client,
        timeout,
        limit,
    )
    # LOCAL MTM QUERIES: fetch relationships directly from the os_workspace_relationship table, then fetch target entities
    expanded_entities = await expand_mtm_local(
        expanded_entities,
        entities_by_concept,
        ontology_rels,
        concepts_to_local_rels,
        metrics,
        client,
        timeout,
        limit,
    )
    # MIXED MTM QUERIES: query via timbr to make the most of JOINs
    expanded_entities = await expand_mtm_mixed(
        expanded_entities,
        entities_by_concept,
        ontology,
        concepts_to_mixed_rels,
        ontology_rels,
        metrics,
        client,
        timeout,
        limit,
    )
    return expanded_entities



import itertools
import math
from typing import Optional, List, Tuple
from pydantic import BaseModel, ConfigDict, Field
from collections import Counter

from presidio_analyzer import AnalyzerEngine, BatchAnalyzerEngine, RecognizerRegistry, AnalysisExplanation, \
    EntityRecognizer, RecognizerResult
from presidio_analyzer.nlp_engine import NlpArtifacts, NlpEngineProvider
import streamlit as st
import nltk
import pandas as pd
from flair.data import Sentence
from flair.models import SequenceTagger

from sumy.parsers.plaintext import PlaintextParser
from sumy.nlp.tokenizers import Tokenizer
from sumy.nlp.stemmers import Stemmer
from sumy.summarizers.lsa import LsaSummarizer
from sumy.summarizers.luhn import LuhnSummarizer
from sumy.utils import get_stop_words

from .language import alpha2_to_language

BASE_ALLOWED_LABELS = ["PERSON", "ORG", "LOC", "NORP", "GPE", "PRODUCT", "DATE", "PHONE", "IP_ADDRESS", "EMAIL", "URL",
                       "CRYPTO", "IBAN", "CREDIT_CARD", "US_SSN", "US_DRIVER_LICENSE", "US_PASSPORT", "MEDICAL_LICENSE"]

PRESIDIO_TO_BASE_ALIASES = {
    "PHONE_NUMBER": "PHONE",
    "EMAIL_ADDRESS": "EMAIL",
    "IBAN_CODE": "IBAN",
    "DRIVER_LICENSE": "US_DRIVER_LICENSE",
    "US_DRIVER_LICENSE": "US_DRIVER_LICENSE",
    "US_DRIVERS_LICENSE": "US_DRIVER_LICENSE",
    "PASSPORT": "US_PASSPORT",
    "CREDIT_CARD": "CREDIT_CARD",
    "URL": "URL",
    "IP_ADDRESS": "IP_ADDRESS",
    "CRYPTO": "CRYPTO",
    "CRYPTO_WALLET": "CRYPTO",
    "CRYPTO_WALLET_ADDRESS": "CRYPTO",
    "DATE_TIME": "DATE",
    "LOCATION": "LOC",
    "ORGANIZATION": "ORG",
}

BASE_TO_RECOGNIZER_EXPANSIONS = {
    "ORG": ["ORG", "ORGANIZATION"],
    "LOC": ["LOC", "LOCATION"],
    "PHONE": ["PHONE", "PHONE_NUMBER"],
    "EMAIL": ["EMAIL", "EMAIL_ADDRESS"],
    "IBAN": ["IBAN", "IBAN_CODE"],
    "US_DRIVER_LICENSE": ["US_DRIVER_LICENSE", "US_DRIVERS_LICENSE", "DRIVER_LICENSE"],
    "US_PASSPORT": ["US_PASSPORT", "PASSPORT"],
    "DATE": ["DATE", "DATE_TIME"],
    "PERSON": ["PERSON"],
    "URL": ["URL"],
    "IP_ADDRESS": ["IP_ADDRESS"],
    "CRYPTO": ["CRYPTO", "CRYPTO_WALLET", "CRYPTO_WALLET_ADDRESS"],
    "CREDIT_CARD": ["CREDIT_CARD"],
    "US_SSN": ["US_SSN"],
    "MEDICAL_LICENSE": ["MEDICAL_LICENSE"],
    "NORP": ["NORP"],
    "GPE": ["GPE"],
    "PRODUCT": ["PRODUCT"],
}

BASE_TO_ONTONOTES_LABELMAP = {"PER": "PERSON"}


class FlairRecognizer(EntityRecognizer):
    ENTITIES = [
        "LOC",
        "PERSON",
        "ORG",
    ]

    DEFAULT_EXPLANATION = "Identified as {} by Flair's Named Entity Recognition"

    CHECK_LABEL_GROUPS = [
        ({"LOC"}, {"LOC", "LOCATION"}),
        ({"PERSON"}, {"PER", "PERSON"}),
        ({"ORG"}, {"ORG", "ORGANIZATION"}),
    ]

    MODEL_LANGUAGES = {
        "en": "flair/ner-english-large",
        "es": "flair/ner-spanish-large",
        "de": "flair/ner-german-large",
        "nl": "flair/ner-dutch-large",
        "multi": "flair/ner-multi",
        "multi-fast": "flair/ner-multi-fast",
    }

    PRESIDIO_EQUIVALENCES = {
        "PER": "PERSON",
        "LOC": "LOC",
        "ORG": "ORG"
    }

    def __init__(
            self,
            model: SequenceTagger = None,
            supported_language: str = "en",
            supported_entities: Optional[List[str]] = None,
            check_label_groups: Optional[Tuple[set, set]] = None,
    ):
        self.check_label_groups = (
            check_label_groups if check_label_groups else self.CHECK_LABEL_GROUPS
        )

        supported_entities = supported_entities if supported_entities else self.ENTITIES
        self.model = model

        super().__init__(
            supported_entities=supported_entities,
            supported_language=supported_language,
            name="Flair Analytics",
        )

    def load(self) -> None:
        pass

    def get_supported_entities(self) -> List[str]:
        return self.supported_entities

    def analyze(self, text: str, entities: List[str], nlp_artifacts: NlpArtifacts = None) -> List[RecognizerResult]:
        results = []

        sentences = Sentence(text)
        self.model.predict(sentences)

        if not entities:
            entities = self.supported_entities

        for entity in entities:
            if entity not in self.supported_entities:
                continue

            for ent in sentences.get_spans("ner"):
                if not self.__check_label(
                        entity, ent.labels[0].value, self.check_label_groups
                ):
                    continue
                textual_explanation = self.DEFAULT_EXPLANATION.format(
                    ent.labels[0].value
                )
                explanation = self.build_flair_explanation(
                    round(ent.score, 2), textual_explanation
                )
                flair_result = self._convert_to_recognizer_result(ent, explanation)

                results.append(flair_result)

        return results

    def build_flair_explanation(self, original_score: float, explanation: str) -> AnalysisExplanation:
        explanation = AnalysisExplanation(
            recognizer=self.__class__.__name__,
            original_score=original_score,
            textual_explanation=explanation,
        )
        return explanation

    def _convert_to_recognizer_result(self, entity, explanation) -> RecognizerResult:
        entity_type = self.PRESIDIO_EQUIVALENCES.get(entity.tag, entity.tag)
        flair_score = round(entity.score, 2)

        flair_results = RecognizerResult(
            entity_type=entity_type,
            start=entity.start_position,
            end=entity.end_position,
            score=flair_score,
            analysis_explanation=explanation,
        )

        return flair_results

    @staticmethod
    def __check_label(
            entity: str, label: str, check_label_groups: Tuple[set, set]
    ) -> bool:
        return any(
            [entity in egrp and label in lgrp for egrp, lgrp in check_label_groups]
        )


def normalize_label(label: str) -> str:
    return PRESIDIO_TO_BASE_ALIASES.get(label, label)


def expand_entities_for_analyzer(entities_list):
    expanded = set()
    for e in entities_list:
        vals = BASE_TO_RECOGNIZER_EXPANSIONS.get(e, [e])
        for v in vals:
            expanded.add(v)
    return list(expanded)


def _sumy__get_best_sentences(sentences, rating, *args, **kwargs):
    from operator import attrgetter
    from sumy.summarizers._summarizer import SentenceInfo

    rate = rating
    if isinstance(rating, dict):
        assert not args and not kwargs
        rate = lambda s: rating[s]
    infos = (SentenceInfo(s, o, rate(s, *args, **kwargs)) for o, s in enumerate(sentences))
    infos = sorted(infos, key=attrgetter("rating"), reverse=True)
    return tuple((i.sentence, i.rating, i.order) for i in infos)


def _sumy__lsa_call(summarizer, document):
    summarizer._ensure_dependecies_installed()
    dictionary = summarizer._create_dictionary(document)
    if not dictionary:
        return ()
    matrix = summarizer._create_matrix(document, dictionary)
    matrix = summarizer._compute_term_frequency(matrix)
    from numpy.linalg import svd as singular_value_decomposition

    u, sigma, v = singular_value_decomposition(matrix, full_matrices=False)
    ranks = iter(summarizer._compute_ranks(sigma, v))
    return _sumy__get_best_sentences(document.sentences, lambda s: next(ranks))


def _sumy__luhn_call(summarizer, document):
    words = summarizer._get_significant_words(document.words)
    return _sumy__get_best_sentences(document.sentences, summarizer.rate_sentence, words)


def get_nltk_tokenizer(language: str) -> Tokenizer:
    nltk_lang = alpha2_to_language(language).lower()

    try:
        nltk.data.find("tokenizers/punkt")
    except LookupError:
        nltk.download("punkt")

    return Tokenizer(nltk_lang)


class NERObject(BaseModel):
    name: str
    label: str
    score: float = 0.0
    start: int
    count: int
    context: str | None = None
    comentions: list[str] = Field(default_factory=list)
    model_config = ConfigDict(extra="allow")

    def __repr__(self):
        return f"NERObject(label={self.label},name={self.name})"


def postprocess_ner(entities: list[NERObject], whitelisted_labels=None, max_entities=None):
    if whitelisted_labels is not None:
        entities = [e for e in entities if e.label in whitelisted_labels]
    entities = sorted(entities, key=lambda x: x.name)
    final_entities = []
    for _, group in itertools.groupby(entities, key=lambda x: x.name):
        group = list(group)
        best_entity = max(group, key=lambda x: x.score * x.count)
        merged_data = {
            "name": best_entity.name,
            "label": best_entity.label,
            "score": best_entity.score,
            "context": best_entity.context,
            "count": sum(e.count for e in group),
            "start": best_entity.start,
        }
        all_fields = best_entity.model_fields.keys()
        for field in all_fields:
            if field in merged_data:
                continue
            values = [getattr(e, field, None) for e in group if getattr(e, field, None) is not None]
            if not values:
                continue
            if isinstance(values[0], list):
                merged_data[field] = list(set(itertools.chain.from_iterable(values or [])))
            else:
                merged_data[field] = getattr(best_entity, field, None)
        final_entities.append(NERObject(**merged_data))
    final_entities = sorted(final_entities, key=lambda x: x.score * x.count, reverse=True)
    if max_entities and len(final_entities) > max_entities:
        final_entities = final_entities[:max_entities]
    return final_entities


def build_presidio_analyzer(language: str, engine_type: str = "spacy", model=None) -> AnalyzerEngine:
    registry = RecognizerRegistry()

    if engine_type == "flair":

        flair_recognizer = FlairRecognizer(
            model=model,
            supported_language=language
        )
        registry.add_recognizer(flair_recognizer)

        default_registry = RecognizerRegistry()
        default_registry.load_predefined_recognizers()

        flair_handled_entities = {"PERSON", "LOC", "ORG"}

        for recognizer in default_registry.recognizers:
            recognizer_entities = set(recognizer.supported_entities) if hasattr(recognizer, 'supported_entities') else set()

            if recognizer_entities and recognizer_entities.issubset(flair_handled_entities):
                continue

            registry.add_recognizer(recognizer)

        return AnalyzerEngine(
            registry=registry,
            supported_languages=[language]
        )

    else:
        registry.load_predefined_recognizers()

        if model is None:
            raise ValueError("SpaCy model name must be provided")

        configuration = {
            "nlp_engine_name": "spacy",
            "models": [{"lang_code": language, "model_name": model}],
        }

        provider = NlpEngineProvider(nlp_configuration=configuration)
        nlp_engine = provider.create_engine()

        return AnalyzerEngine(
            nlp_engine=nlp_engine,
            registry=registry,
            supported_languages=[language],
        )


def analyze_column_sample(column_values: pd.Series, analyzer: AnalyzerEngine, language: str,
                          entities: Optional[List[str]], score_threshold: float) -> Optional[str]:
    sample_values = column_values.dropna().head(50)

    if sample_values.empty:
        return None

    entity_counter = Counter()

    for value in sample_values:
        text = str(value).strip()

        if not text:
            continue

        results = analyzer.analyze(
            text=text,
            language=language,
            entities=(expand_entities_for_analyzer(entities) if entities else None)
        )

        for result in results:
            if result.score >= score_threshold:
                entity_counter[normalize_label(result.entity_type)] += 1

    if not entity_counter:
        return None

    most_common = entity_counter.most_common(1)[0]
    total_detections = sum(entity_counter.values())

    if most_common[1] > total_detections * 0.5:
        return most_common[0]

    return most_common[0] if entity_counter else None


def analyze_dataframe_optimized(df: pd.DataFrame, analyzer: AnalyzerEngine, language: str,
                                entities: Optional[List[str]] = None, score_threshold: float = 0.5) -> List[NERObject]:
    ner_objects = []

    for column_name in df.columns:
        entity_type = analyze_column_sample(
            df[column_name],
            analyzer,
            language,
            entities,
            score_threshold
        )

        if entity_type:
            for idx, value in df[column_name].dropna().items():
                text = str(value).strip()

                if text:
                    ner_objects.append(NERObject(
                        name=text[:100],
                        label=entity_type,
                        score=0.9,
                        start=0,
                        count=1,
                        context=text[:100]
                    ))

    return ner_objects


def compute_ner_presidio(
        text,
        language,
        analyzer,
        entities=None,
        score_threshold=0.5,
        context_width=150,
        with_comentions=True,
        with_context=True,
        batch_size=32,
        n_process=4
):
    if isinstance(text, pd.DataFrame):
        if len(text) >= 100:
            return analyze_dataframe_optimized(text, analyzer, language, entities, score_threshold)

        else:
            texts = []

            for col in text.columns:
                for idx, value in text[col].dropna().items():
                    text_value = str(value).strip()

                    if text_value:
                        texts.append(text_value)

            text = "\n".join(texts)

    elif isinstance(text, list):
        batch_analyzer = BatchAnalyzerEngine(analyzer_engine=analyzer)

        results_generator = batch_analyzer.analyze_iterator(
            texts=text,
            language=language,
            batch_size=batch_size,
            n_process=n_process,
            entities=(expand_entities_for_analyzer(entities) if entities else None),
        )

        all_results = list(results_generator)
        ner_objects = []

        for text_item, results in zip(text, all_results):
            for result in results:
                if result.score >= score_threshold:
                    context_start = max(0, result.start - 30)
                    context_end = min(len(text_item), result.end + 30)
                    context = text_item[context_start:context_end] if with_context else None

                    ner_objects.append(NERObject(
                        name=text_item[result.start:result.end],
                        label=normalize_label(result.entity_type),
                        score=float(result.score),
                        start=int(result.start),
                        count=1,
                        context=context
                    ))

        return ner_objects

    results = analyzer.analyze(
        text=text,
        language=language,
        entities=(expand_entities_for_analyzer(entities) if entities else None)
    )

    ner_objects = []

    for result in results:
        if result.score >= score_threshold:
            context_start = max(0, result.start - math.floor(context_width / 2))
            context_end = min(len(text), result.end + math.ceil(context_width / 2))
            context = text[context_start:context_end] if with_context else None

            ner_objects.append(NERObject(
                name=text[result.start:result.end],
                label=normalize_label(result.entity_type),
                score=float(result.score),
                start=int(result.start),
                count=1,
                context=context
            ))

    if with_comentions:
        for i in range(len(ner_objects)):
            entity = ner_objects[i]
            comentions = [
                ner_objects[j].name
                for j in range(len(ner_objects))
                if j != i and abs(ner_objects[j].start - entity.start) < math.ceil(context_width / 2)
            ]
            ner_objects[i].comentions = comentions

    return ner_objects


def get_extractive_summary(text, language, max_chars, fast=False, with_scores=False):
    tokenizer = get_nltk_tokenizer(language)
    stemmer = Stemmer(language)
    parser = PlaintextParser.from_string(text, tokenizer)
    if fast:
        summarizer = LuhnSummarizer(stemmer)
        summarizer.stop_words = get_stop_words(language)
        scored_sentences = iter(_sumy__luhn_call(summarizer, parser.document))
    else:
        summarizer = LsaSummarizer(stemmer)
        summarizer.stop_words = get_stop_words(language)
        scored_sentences = iter(_sumy__lsa_call(summarizer, parser.document))
    summary = []
    summary_chars = 0
    summary_chars_penultimate = 0
    while summary_chars < max_chars:
        try:
            next_sentence = next(scored_sentences)
            summary.append(next_sentence)
            summary_chars_penultimate = summary_chars
            summary_chars += len(" " + next_sentence[0]._text)
        except StopIteration:
            break
    summary = sorted(summary, key=lambda x: x[2])
    summary = [(sentence[0]._text, sentence[1]) for sentence in summary]
    if summary_chars > max_chars:
        summary[-1] = (
            summary[-1][0][: max_chars - summary_chars_penultimate],
            summary[-1][1],
        )
    if not with_scores:
        summary = " ".join([s[0] for s in summary])
    else:
        min_score = min([s[1] for s in summary]) if summary else 0
        max_score = max([min_score] + [s[1] for s in summary])
        score_range = 1 if min_score == max_score else (max_score - min_score)
        summary = [(s[0], (s[1] - min_score) / score_range) for s in summary]
    return summary


def ner_pipe(
        text,
        language,
        model,
        engine_type="spacy",
        fast=False,
        compression_ratio="auto",
        with_comentions=True,
        with_context=True,
        entities=None,
        score_threshold=0.5,
        batch_size=32,
        n_process=4
):
    analyzer = build_presidio_analyzer(
        language=language,
        engine_type=engine_type,
        model=model,
    )

    if isinstance(text, pd.DataFrame):
        ner = compute_ner_presidio(
            text,
            language,
            analyzer,
            entities,
            score_threshold,
            with_comentions=with_comentions,
            with_context=with_context,
            batch_size=batch_size,
            n_process=n_process
        )
    else:
        if compression_ratio == "auto":
            compression_ratio = max(1.0, len(text) / 15000) if fast else 1.0

        if compression_ratio > 1.0:
            sentences = get_extractive_summary(text, language, int(len(text) / compression_ratio), fast=fast,
                                               with_scores=True)
            text = " ".join([s[0] for s in sentences])

        ner = compute_ner_presidio(
            text,
            language,
            analyzer,
            entities,
            score_threshold,
            with_comentions=with_comentions,
            with_context=with_context,
            batch_size=batch_size,
            n_process=n_process
        )

    return ner


def get_ner_handler(
        language,
        model,
        engine_type="spacy",
        fast=False,
        entities=None,
        score_threshold=0.5,
        batch_size=32,
        n_process=4
):
    try:
        get_nltk_tokenizer(language)
    except LookupError:
        language = "en"

    return lambda text, compression_ratio="auto", with_comentions=True, with_context=True: ner_pipe(
        text,
        language,
        model,
        engine_type,
        fast,
        compression_ratio,
        with_comentions,
        with_context,
        entities,
        score_threshold,
        batch_size,
        n_process
    )


@st.cache_resource
def get_cached_ner_handler(language, model):
    return get_ner_handler(language, model)
